/*********************************************************************
    Copyright © 2002-2022 by Ruud Harmsen.

    Virtrend calculates the virtual yield of an investment. See
    https://rudhar.com/sfreview/virtrend and
    https://rudhar.com/sfreview/virtrend/en.htm for explanation.

    Virtrend is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    Virtrend is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Siworin.  If not, see <https://www.gnu.org/licenses/>.
  ********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

/* [nl] Virtueel (effectief) rendement van een aandelenbelegging,
	alsof het een spaarrekening was.
   [en] Virtual (effective ) yield of a stock investment or similar,
	as if it were a simple savings account.
   Copyright 2002-2004, 2015, 2022 Ruud Harmsen

   26 Aug 2015: Better input checks and diagnostics.
    1 Dec 2022: More measures for robustness against bad input.
    7 Dec 2022: More tests.
    7 Dec 2022: Fixed non-convergence in case of a negative starting
                value, e.g. to simulate a loan instead of a savings
                account.
   */

#define MAXTRIES 85
struct pointintime_s
{
   char DateFrom[8+1];
   int YearFrom;
   int InterestNow;
   long DaysSince1970;
   double CumulAmount;
   double AddAmount;
   double CurrencyRate;
   double InterestDaysBasic;
   int SeqNumOninput;
};

int EltsAllocated = 0;

static struct pointintime_s *Readfile (
   struct pointintime_s *D,
   char *Infile,
   int *pEltsAllocated,
   int *pInputErrorsFound);
static int mkdays (struct pointintime_s *d);
static int RudimentaryDateCheck (char *line);

static struct pointintime_s *SortAndadd (
   struct pointintime_s *D,
   int *pEltsAllocated);
static int CompElt (const void *a, const void *b);

static struct pointintime_s *Calculate (
   struct pointintime_s *D,
   int EltsAllocated);

static double CalcEndValue (
   struct pointintime_s *D,
   int EltsAllocated,
   double Percentage);

static double NewPercentage (
   double OldPercentage,
   double CalqedValue,
   double ActualValue,
   int posneg);

static int ShowTable (struct pointintime_s *d, int EltsAllocated, int sourceline);

int main (int argc, char **argv)
{
   struct pointintime_s *D = NULL;
   int retval = 0;
   int InputErrorsFound = 0;

   if (argc <= 1)
   {
      D = Readfile(D, "virtrend.in", &EltsAllocated, &InputErrorsFound);
      if (!D)
      {
         fprintf(stderr, "No data file\n");
         exit(2);
      }
   }
   else
   {
      char **f = argv;
      for (f = f + 1; *f; f++)
      {
         D = Readfile(D, *f, &EltsAllocated, &InputErrorsFound);
         if (!D)
         {
            exit(1);
         }
      }
   }

   if (EltsAllocated <= 0)
   {
      fprintf(stdout,
         "%s: Nothing to do, no data. Bye!\n", argv[0]);
      retval = 1;
   }
   else if (InputErrorsFound)
   {
      fprintf(stdout,
         "%s: Errors in data, cannot do calculations, Bye!\n", argv[0]);
      retval = 1;
   }
   else
   {
      D = SortAndadd(D, &EltsAllocated);
      Calculate(D, EltsAllocated);
   }
   if (D) free(D);

   return retval;
}

#define AllocateEachTime 3
static struct pointintime_s *Readfile (
   struct pointintime_s *D,
   char *Infile,
   int *pEltsAllocated,
   int *pInputErrorsFound)
{
   FILE *fp;
   char line[256];
   static int NextElt = 0;
   static int linenum = 0;
   struct pointintime_s last;

   memset(&last, '\0', sizeof last);

   fp = fopen(Infile, "rt");
   if (!fp)
   {
      Infile = "virtrend.in";
      fp = fopen(Infile, "rt");
   }
   if (!fp)
      return NULL;
   else
      printf("Reading file %s\n", Infile);

   if (!D)
   {
      #ifdef DEBUG
      fprintf(stdout, "Progline %d, inputline %d\n",  __LINE__, linenum);
      fprintf(stdout, "%s\n", line);
      #endif
      D = malloc(AllocateEachTime * sizeof (*D));
      if (!D)
         return NULL;
      *pEltsAllocated = AllocateEachTime;
      NextElt = 0;
   }
   else if (NextElt > 0)
   {
      last = D[NextElt - 1];
      NextElt--;
   }

   while (fgets(line, sizeof line, fp) == line)
   {
      char *hashpos = strchr(line, '#');
      size_t clen;
      char c;

      if (hashpos) clen = strlen(hashpos);
      while (hashpos &&
             clen + 1 == (sizeof line) - (hashpos - line) &&
             hashpos[clen - 1] != '\n' &&
             !feof(fp))
      {
         #ifdef DEBUG
            fprintf(stdout, "Line nr. %d longer than buffer. Line is:\n",
               linenum + 1);
            fprintf(stdout, "\"%s\"\n", line);
         #endif
         /* Keep on reading if line longer than line buffer */
         fgets(hashpos + 1, (sizeof line) - (hashpos - line) - 1, fp);
         clen = strlen(hashpos);
      }

      linenum++;
      #ifdef DEBUG
         fprintf(stdout,
            "Progline %d: Program just read line nr. %d. Line is:\n",
            __LINE__, linenum);
         fprintf(stdout, "%s\n", line);
      #endif

      c = line[strspn(line, " \t\r\n")];
      if (c == '\0' || c == '#')
      {
         #ifdef DEBUG
            fprintf(stdout, "Skipping empty line or comment\n\n");
         #endif
         continue;
      }

      if (NextElt >= *pEltsAllocated)
      {
         D = realloc(D, sizeof (struct pointintime_s) *
                (*pEltsAllocated + AllocateEachTime));
         if (!D)
         {
            fclose(fp);
            return NULL;
         }
         *pEltsAllocated += AllocateEachTime;
      }
      #ifdef DEBUG
         fprintf(stdout, "Progline %d, inputline %d\n",
            __LINE__, linenum);
         fprintf(stdout, "%s\n", line);
      #endif
      if (RudimentaryDateCheck(line))
      {
         fprintf(stdout, "Invalid date \"%-8.8s\" at line %d.\n",
            line, linenum);
         *pInputErrorsFound = 1;
         continue;
      }
      D[NextElt].CurrencyRate = 1.0;
      sscanf(line, "%s%lg%lg",
         D[NextElt].DateFrom, &D[NextElt].AddAmount,
                              &D[NextElt].CurrencyRate);
      D[NextElt].AddAmount /= D[NextElt].CurrencyRate;
      D[NextElt].CumulAmount = 0;
      mkdays(D + NextElt);
      D[NextElt].InterestNow = 0;
      D[NextElt].SeqNumOninput = NextElt;

      NextElt++;
   }

   fclose(fp);

   D[NextElt - 1].AddAmount += last.AddAmount;
   if (strcmp(last.DateFrom, D[NextElt - 1].DateFrom) > 0)
   {
      strcpy(D[NextElt - 1].DateFrom, last.DateFrom);
      mkdays(D + NextElt - 1);
   }

   *pEltsAllocated = NextElt;

   return D;
}

static int mkdays (struct pointintime_s *d)
{
   struct tm date;
   char yr[4+1];
   char mn[2+1];
   char da[2+1];
   time_t secs;

   sprintf(yr, "%4.4s", d->DateFrom);
   sprintf(mn, "%2.2s", d->DateFrom + 4);
   sprintf(da, "%2.2s", d->DateFrom + 6);

   date.tm_year = atoi(yr) - 1900;
   date.tm_mon  = atoi(mn) - 1;
   date.tm_mday = atoi(da);
   date.tm_hour = date.tm_min = date.tm_sec  = 0;
   secs = mktime(&date);

   d->YearFrom = atoi(yr);
   d->DaysSince1970 = secs / 60.0 / 60.0 / 24;

   return 0;
}

static int RudimentaryDateCheck (char *line)
{
   char *s = line;
   int i = 8;

   while (*s && isspace(*s))
      s++;

   while (--i > 0)
   {
      if (!isdigit(*s))
         return 1;
   }
   return 0;
}

static struct pointintime_s *SortAndadd (
   struct pointintime_s *D,
   int *pEltsAllocated)
{
   struct pointintime_s *ld, *nd;
   struct pointintime_s *NewD;
   int ExtraRecs;
   int i;

   /* Sort all the entries by ascending date, so it
      doesn't matter whether or not the user already
      did that too. Extra advantage: merging of files
      is easier. */
   qsort(D, *pEltsAllocated, sizeof (*D), CompElt);

   /* Insert extra records, one each year, for assigning
      the virtual yearly interest */
   ExtraRecs = D[*pEltsAllocated - 1].YearFrom - D[0].YearFrom;

   #ifdef DEBUG
   fprintf(stdout, "Progline %d\n",  __LINE__);
   #endif

   NewD = malloc((*pEltsAllocated + ExtraRecs) * sizeof (*D));

   for (ld = D, nd = NewD; ld - D < *pEltsAllocated; ld++, nd++)
   {
      int year;
      *nd = *ld;
      if (ld - D == *pEltsAllocated - 1)
      {
         i = 0;
      }
      else
      {
         i = ld[1].YearFrom - ld[0].YearFrom;
         year = ld[0].YearFrom + 1;
      }
      while (i--)
      {
         ++nd;
         sprintf(nd->DateFrom, "%04d0101", year++);
         mkdays(nd);
         nd->CumulAmount   = nd->AddAmount = 0.0;
         nd->CurrencyRate  = 1.0;
         nd->InterestNow   = 1;
         nd->SeqNumOninput = 0;
      }
   }
   *pEltsAllocated += ExtraRecs;
   NewD[*pEltsAllocated - 1].InterestNow = 1;

   free(D);
   return NewD;
}

static int CompElt (const void *a, const void *b)
{
   struct pointintime_s *A, *B;

   A = (struct pointintime_s *)a;
   B = (struct pointintime_s *)b;

   if (strcmp(A->DateFrom, B->DateFrom) == 0)
      return A->SeqNumOninput - B->SeqNumOninput;

   return strcmp(A->DateFrom, B->DateFrom);
}

static struct pointintime_s *Calculate (
   struct pointintime_s *D,
   int EltsAllocated)
{
   struct pointintime_s *ld;
   double bals = 0;
   double Percentage;
   double ActualValue, CalqedValue;
   long tryze;
   int p;

   for (ld = D; ld - D < EltsAllocated; ld++)
   {
      bals += ld->AddAmount;
      ld->CumulAmount = bals;
   }

   for (ld = D; ld - D < EltsAllocated - 1; ld++)
   {
      ld->InterestDaysBasic = ld->CumulAmount / 100.0 *
         (ld[1].DaysSince1970 - ld->DaysSince1970);
   }

   #ifdef DEBUG
   ShowTable (D, EltsAllocated, __LINE__);
   #endif

   /* In the last row, the field AddAmount is to be interpreted as the
      actual value of the shares now (i.e. the date specified). */
   ActualValue = D[EltsAllocated - 1].AddAmount;

   /* Now try various interest percentages, which, if this were a
      simple savings account, would have produced the same
      end value */
   Percentage = 4.0;
   CalqedValue = CalcEndValue(D, EltsAllocated, Percentage);

   for (tryze = 1; tryze < MAXTRIES && fabs(ActualValue - CalqedValue) > 0.01; tryze++)
   {
      Percentage = NewPercentage(
         Percentage, CalqedValue, ActualValue, D[0].CumulAmount < 0 ? -1 : 1);
      CalqedValue = CalcEndValue(D, EltsAllocated, Percentage);
      #ifdef DEBUG2
      fprintf(stdout, "Percentage %13.9f, CalqedValue %11.2f ActualValue %11.2f\n",
         Percentage, CalqedValue, ActualValue);
      #endif
   }

   printf("After %ld attempts: %6.3f%%\n", tryze, Percentage);
   for (p = -4; p < 9; p++)
   {
      printf("Result compared with %2d%%: %9.2f\n",
         p, ActualValue - CalcEndValue(D, EltsAllocated, p));
   }
   return D;
}

static double CalcEndValue (
   struct pointintime_s *d,
   int EltsAllocated,
   double Percentage)
{
   struct pointintime_s *ld;
   double CumulInterestDays = 0;
   double ExtraInterestDays;
   double PayedInterest = 0;

   for (ld = d; ld - d < EltsAllocated; ld++)
   {
      if (ld->InterestNow)
      {
         PayedInterest += CumulInterestDays * Percentage / 365.25;
         CumulInterestDays = 0;
      }
      CumulInterestDays += ld->InterestDaysBasic;

      /* [nl] Rente op rente, samengestelde interest.
	 [en] Compound interest or composite interest. */
      ExtraInterestDays = PayedInterest / 100.0 *
         (ld[1].DaysSince1970 - ld[0].DaysSince1970);
      CumulInterestDays += ExtraInterestDays;
   }
   return d[EltsAllocated - 2].CumulAmount + PayedInterest;
}

static double NewPercentage (
   double OldPercentage,
   double CalqedValue,
   double ActualValue,
   int posneg)
{
   double Percentage;
   static float Steps[] = {
          3,            1,
          0.3,          0.1,
          0.03,         0.01,
          0.003,        0.001,
          0.0003,       0.0001,
          0.00003,      0.00001,
          0.000003,     0.000001,
          0.0000003,    0.0000001,
          0.00000003,   0.00000001,
          0.000000003,  0.000000001,
          0.0000000003, 0.0000000001,
          };
   static const int numSteps = sizeof Steps / sizeof Steps[0];
   static int StepSize = 0;
   int Direction;
   static int PrevDirection = 1;

   Direction = CalqedValue > ActualValue ? -1 : +1;
   Direction *= posneg;

   if (Direction * PrevDirection == -1)
      StepSize++; /* Larger index means smaller steps */
   if (StepSize >= numSteps)
      StepSize = numSteps - 1;
   else if (StepSize < 0)
      StepSize = 0;
   Percentage = OldPercentage + Steps[StepSize] * Direction;

   PrevDirection = Direction;

   return Percentage;
}

#if defined DEBUG || defined DEBUG2
static int ShowTable (struct pointintime_s *d, int EltsAllocated, int sourceline)
{
   struct pointintime_s *ld;

   fprintf(stdout, "==== Called from prog.line %d =====\n", sourceline);
   for (ld = d; ld - d < EltsAllocated; ld++)
   {
      fprintf(stdout, "%2d) %s %1d %9.2f %9.2f %9.2f\n", (int)(ld - d),
         ld->DateFrom, ld->InterestNow,
         ld->CumulAmount * ld->CurrencyRate,
         ld->AddAmount   * ld->CurrencyRate,
         ld->InterestDaysBasic);
   }
   return 0;
}
#endif
